import json
from pyproj import CRS

from .schema import Schema


class Info(object):

    def __init__(self, data):
        self.data = json.loads(data)

    def __len__(self):
        return int(self.data['points'])

    def get_schema(self):
        return Schema(self.data['schema'])
    schema = property(get_schema)

    def get_span(self):
        return int(self.data['span'])
    span = property(get_span)

    def get_version(self):
        return self.data['version']
    version = property(get_version)

    def get_bounds(self):
        return self.data['bounds']
    bounds = property(get_bounds)

    def get_conforming(self):
        return self.data['boundsConforming']
    conforming = property(get_conforming)

    def get_datatype(self):
        return self.data['dataType']
    datatype = property(get_datatype)

    def get_hierarchytype(self):
        return self.data['hierarchyType']
    hierarchytype = property(get_hierarchytype)

    def get_srs(self):
        wkt = self.data['srs']['wkt']
        crs = CRS.from_user_input(wkt)
        return crs
    srs = property(get_srs)



#
# Endpoint module
#
import json

import aiohttp
import aiofiles
import asyncio

from .pool import TaskPool


class Driver(object):
    def __init__(self, root, concurrency=1):
        self.root = root
        self.parts = []
        self.concurrency = concurrency


class Http(Driver):
    def __init__(self, root, query=None):
        super(Http, self).__init__(root)
        self.query = query

    async def download(self, session, url):
        async with session.get(url) as response:
            return await response.read()

    async def get(self, part, session=None, tpool=None):
        url = self.root + part
        if self.query is not None:
            url += '?' + self.query
        if tpool:
            return await tpool.put(self.download(session, url))
        if session:
            async with session.get(url) as response:
                return await response.read()
        else:
            async with aiohttp.ClientSession() as session:
                async with session.get(url) as response:
                    return await response.read()

    def stage(self, part):
        self.parts.append(part)

    async def bulk(self):
        connector = aiohttp.TCPConnector(limit=None)
        async with aiohttp.ClientSession(connector=connector) as session, TaskPool(
            self.concurrency
        ) as tasks:
            for part in self.parts:
                await tasks.put(self.download(session, self.root + part))

        return tasks


class File(Driver):
    def __init__(self, root):
        super(File, self).__init__(root)

    async def get(self, part, session=None, tpool=None):
        url = self.root
        if part:
            url = url + part

        async with aiofiles.open(url, "rb") as d:
            return await d.read()


class Endpoint(object):
    def __init__(self, root, query=None):
        self.root = root
        self.query = query

        if root.startswith("http://") or root.startswith("https://"):
            self.remote = True
            self.driver = Http(root, query)
        else:
            self.remote = False
            self.driver = File(root)

    def get(self, part):
        loop = asyncio.get_event_loop()
        o = loop.run_until_complete(self.driver.get(part))
        return o

    async def aget(self, part=None, session=None, tpool=None):
        return await self.driver.get(part, session, tpool)



import copy

class Bounds(object):
    def __init__(self, *args):
        self.coords = args
    def __repr__(self):
        return repr(self.coords)


class Key(object):
    def __init__(self, k=None):
        self.ids = [0, 0, 0]
        self.coords = [0.0 for i in range(6)]
        self.d = 0
        self.count = 0

        if k:
            if isinstance(k, Key):
                self.d = k.d
                self.ids = k.ids
                self.coords = k.coords

            elif len(k):
                self.d, x, y, z = [int(t) for t in k.split('-')]

                self.ids[0] = x
                self.ids[1] = y
                self.ids[2] = z

    def id(self):
        return '%d-%d-%d-%d' % (self.d, self.ids[0], self.ids[1], self.ids[2])

    def __repr__(self):
        minx = self.coords[0]; miny = self.coords[1]; minz = self.coords[2]
        maxx = self.coords[3]; maxy = self.coords[4]; maxz = self.coords[5]

        return "id: %s, box3d(%.3f %.3f %.3f %.3f %.3f %.3f) %d" % (self.id(), minx, miny, minz, maxx, maxy, maxz, self.count)

    def contains(self, other):
        return  self.coords[0] <= other.coords[0] and self.coords[3] <= other.coords[3] and \
                self.coords[1] <= other.coords[1] and self.coords[4] <= other.coords[4] and \
                self.coords[2] <= other.coords[2] and self.coords[5] <= other.coords[5]

    def overlaps(self, other):
        return  self.coords[0] <= other.coords[3] and self.coords[3] >= other.coords[0] and \
                self.coords[1] <= other.coords[4] and self.coords[4] >= other.coords[1] and \
                self.coords[2] <= other.coords[5] and self.coords[5] >= other.coords[2]

    def bisect(self, direction):

        key = copy.deepcopy(Key(self))

        key.d = key.d + 1

        def step(i):
            key.ids[i] = key.ids[i] * 2

            mid = key.coords[i] + (key.coords[i+3] - key.coords[i])/2.0

            positive = bool(direction & (int(1) << i))
            if positive:
                key.coords[i] = mid
                key.ids[i] = key.ids[i] + 1
            else:
                key.coords[i + 3] = mid

        for i in range(3):
            step(i)

        return key




import json
import unittest

from ept.endpoint import Endpoint

class TestRemoteEndpoint(unittest.TestCase):

    def setUp(self):
        self.e = Endpoint('https://raw.githubusercontent.com/PDAL/data/master/entwine/data/ept-star/')

    def test_info(self):
        self.assertEqual(self.e.remote, True)

    def test_fetch(self):
        d = self.e.get('/ept.json').replace(b"\r", b"")
        self.assertEqual(len(d), 2192)

class TestQueryParamEndpoint(unittest.TestCase):

    def setUp(self):
        self.e = Endpoint('https://httpbin.org', 'q=42')

    def test_info(self):
        self.assertEqual(self.e.remote, True)
        self.assertEqual(self.e.query, 'q=42')

    def test_fetch(self):
        j = json.loads(self.e.get('/anything'))
        # Our query parameters are parsed out into response.args.q, make sure
        # that they are forwarded properly.
        self.assertEqual(j['args']['q'], '42')

class TestLocalEndpoint(unittest.TestCase):

    def setUp(self):
        self.e = Endpoint('test/ept-star')

    def test_info(self):
        self.assertEqual(self.e.remote, False)

    def test_fetch(self):
        d = self.e.get('/ept.json').replace(b"\r", b"")
        self.assertEqual(len(d), 2192)


if __name__ == '__main__':
    unittest.main()



import unittest

from ept.ept import EPT
from ept.hierarchy import Bounds

class TestEPT(unittest.TestCase):

    def setUp(self):
        self.e = EPT('test/ept-star')
#        self.e = EPT('http://entwine.io/data/ept-star')
#        self.e = Endpoint('https://raw.githubusercontent.com/PDAL/data/master/entwine/data/ept-star/')

    def test_info(self):
        self.assertEqual(len(self.e.info), 518862)
        self.assertEqual(self.e.info.span, 128)
        self.assertEqual(self.e.info.version, '1.0.0')
        self.assertEqual(self.e.info.datatype,'laszip')
        self.assertEqual(self.e.info.hierarchytype,'json')
#
    def test_srs(self):
        self.assertEqual(self.e.info.srs.is_projected, True)
        self.assertEqual(self.e.info.srs.datum.name, 'World Geodetic System 1984')
        self.assertEqual(self.e.info.srs.name, 'WGS 84 / Pseudo-Mercator')
        self.assertEqual(self.e.info.srs.to_epsg(), 3857)
#
    def test_bounds(self):
        self.assertAlmostEqual(self.e.info.bounds[0], 515363)
        self.assertAlmostEqual(self.e.info.bounds[1], 4918339)
        self.assertAlmostEqual(self.e.info.bounds[2], 2309)
        self.assertAlmostEqual(self.e.info.bounds[3], 515407)
        self.assertAlmostEqual(self.e.info.bounds[4], 4918383)
        self.assertAlmostEqual(self.e.info.bounds[5], 2353)
#
    def test_conforming(self):
        self.assertAlmostEqual(self.e.info.conforming[0], 515368)
        self.assertAlmostEqual(self.e.info.conforming[1], 4918340)
        self.assertAlmostEqual(self.e.info.conforming[2], 2322)
        self.assertAlmostEqual(self.e.info.conforming[3], 515402)
        self.assertAlmostEqual(self.e.info.conforming[4], 4918382)
        self.assertAlmostEqual(self.e.info.conforming[5], 2339)
#
    def test_schema(self):
        s = self.e.info.schema
        self.assertEqual(len(s), 14)
#
        dtype = s.dtype
        self.assertEqual(dtype.itemsize, 38)
#
    def test_count(self):
        e = EPT('test/ept-star')
        self.assertEqual(e.count(), 518862)
        self.assertEqual(len(e.overlaps_dict),65)
#
    def test_boundedCount(self):
        e = EPT('test/ept-star')
        b = Bounds(515380, 4918350, 2320, 515400, 4918370, 2325)
        e.queryBounds = b
        self.assertEqual(e.count(), 208375)
#        self.assertEqual(e.count(), 45930) # real count for this box
#
    def test_boundedDepth(self):
        e = EPT('test/ept-star')
#        e = EPT('http://entwine.io/data/ept-star')
#
        e.depthEnd = 3
        self.assertEqual(e.count(), 303955)
        self.assertEqual(len(e.overlaps_dict),27)
#
    def test_boundedResolution(self):
        e = EPT('test/ept-star')
#        e = EPT('http://entwine.io/data/ept-star')
#
        e.queryResolution = 0.1
        self.assertEqual(e.count(), 303955)

    def test_data(self):
        e = EPT('test/ept-star')
#        e = EPT('http://entwine.io/data/ept-star')
        d = e.data()
        print (d[0])


if __name__ == '__main__':
    unittest.main()



import unittest

from ept.info import Info
from ept.hierarchy import Key
#from ept.hierarchy  import Key, Box

class TestKey(unittest.TestCase):

    def test_parse(self):
        k = Key()
        self.assertEqual(k.d, 0)

        with self.assertRaises(ValueError):
            k = Key('d-0-0-0')

        with self.assertRaises(ValueError):
            k = Key('0-0-0-0-0')

        with self.assertRaises(ValueError):
            k = Key('0-0')

    def test_id(self):
        k = Key('1-2-3-4')
        self.assertEqual(k.ids[0], 2)
        self.assertEqual(k.ids[1], 3)
        self.assertEqual(k.ids[2], 4)
        with self.assertRaises(IndexError):
            k.ids[3]


    def test_id_setting(self):
        k = Key('1-2-3-4')

        self.assertEqual(k.ids[0], 2)
        k.ids[0] = k.ids[0]+1
        self.assertEqual(k.ids[0], 3)
        self.assertEqual(k.id(), '1-3-3-4')

    def test_equal(self):
        k = Key('1-2-3-4')
        self.assertEqual(k.id(), '1-2-3-4')
        k = Key(k.id())
        self.assertEqual(k.id(), '1-2-3-4')

    def test_bounds(self):
        k = Key()
        k.coords = [
                515363,
                4918339,
                2309,
                515407,
                4918383,
                2353
        ]

        for direction in range(8):
            k.bisect(direction)

    def test_hash(self):
        k = Key()
        d = {k, 42}


if __name__ == '__main__':
    unittest.main()



#
# Task Pool module
#
import asyncio


# Adapted from https://medium.com/@cgarciae/making-an-infinite-number-of-requests-with-python-aiohttp-pypeln-3a552b97dc95
class TaskPool(object):
    """TaskPool for doing lots of stuff with a semaphore"""

    def __init__(self, workers):
        self._semaphore = asyncio.Semaphore(workers)
        self._tasks = set()
        self.data = {}

    async def put(self, coro):
        await self._semaphore.acquire()

        task = asyncio.ensure_future(coro)

        # cache coro's args with task-id
        args = coro.cr_frame.f_locals
        key_id = task.get_name()

        self.data[key_id] = {"args": args}
        self._tasks.add(task)
        task.add_done_callback(self._on_task_done)

    def _on_task_done(self, task):
        k = task.get_name()
        try:
            self.data[k]["result"] = task.result()
        except:
            self.data[k]["result"] = None
        self.data[k]["exception"] = task.exception()
        self._tasks.remove(task)
        self._semaphore.release()

    async def join(self):
        await asyncio.gather(*self._tasks, return_exceptions=True)

    async def __aenter__(self):
        return self

    def __aexit__(self, exc_type, exc, tb):
        return self.join()



#!/usr/bin/env python
#
#
import logging
import sys
from distutils.version import StrictVersion

from setuptools import setup

logging.basicConfig()
log = logging.getLogger(__file__)

# python -W all setup.py ...
if "all" in sys.warnoptions:
    log.level = logging.DEBUG

# Get the version from the ept module
module_version = None
with open("ept/__init__.py", "r") as fp:
    for line in fp:
        if line.startswith("__version__"):
            module_version = StrictVersion(line.split("=")[1].strip().strip("\"'"))
            break

if not module_version:
    raise ValueError("Could not determine ept's version")

# Handle UTF-8 encoding of certain text files.
open_kwds = {}
if sys.version_info >= (3,):
    open_kwds["encoding"] = "utf-8"

with open("VERSION.txt", "w", **open_kwds) as fp:
    fp.write(str(module_version))

with open("README.md", "r", **open_kwds) as fp:
    readme = fp.read()

with open("CHANGES.md", "r", **open_kwds) as fp:
    changes = fp.read()

long_description = readme + "\n\n" + changes

setup_args = dict(
    name="ept-python",
    version=str(module_version),
    install_requires=[
        "laspy>=2.0.3",
        "lazrs>=0.3.1",
        "pyproj>=3.2.0",
        "numpy>=1.21",
        "aiohttp==3.7.4", # https://github.com/aio-libs/aiohttp/issues/5394
        "aiofiles>=0.7.0",
        "requests>=2.26.0",
    ],
    description="",
    license="BSD",
    keywords="point cloud data organization",
    author="Howard Butler",
    author_email="howard@hobu.co",
    maintainer="Howard Butler",
    maintainer_email="howard@hobu.co",
    url="https://github.com/hobu/ept-python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    test_suite="test",
    include_package_data=True,
    packages=["ept"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: GIS",
    ],
)
setup(**setup_args)



#
# EPT module
#
import os
import json
from urllib.parse import urljoin, urlsplit

import aiohttp
import asyncio
import numpy

from .info import Info
from .hierarchy import Key, Bounds
from .endpoint import Endpoint
from .pool import TaskPool
from .laz import LAZ


class EPT(object):
    def __init__(self, url, bounds=None, queryResolution=None):
        query = None
        if ('?' in url):
            [url, query] = url.split('?', 1)

        if url.endswith("/"):
            url = url[:-1]

        if url.endswith(".json"):
            # gave us path to EPT root
            p = urlsplit(url)
            path = os.path.dirname(p.path)
            url = urljoin(url, path, "/")

        self.query = query
        self.root_url = url
        self.key = Key()
        self.overlaps_dict = {}
        self.depthEnd = None
        self.queryResolution = queryResolution
        self.queryBounds = bounds
        self.endpoint = Endpoint(self.root_url, self.query)
        self.info = self.get_info()
        self.computedDepth = False

    def as_laspy(self, strictbounds=True):
        """
        Method to return a single LasData object for an ept query.
        """

        def filter(las, xmin, ymin, zmin, xmax, ymax, zmax):
            return las.points.array[
                (las.x >= xmin)
                & (las.x < xmax)
                & (las.y >= ymin)
                & (las.y < ymax)
                & (las.z >= zmin)
                & (las.z < zmax)
            ]

        las_objects = self.data()
        if las_objects:
            las = las_objects[0].las
            if strictbounds:
                las.points.array = numpy.concatenate(
                    [filter(l.las, *self.queryBounds.coords) for l in las_objects]
                )
            else:
                las.points.array = numpy.concatenate(
                    [l.las.points.array for l in las_objects]
                )
        else:
            las = None

        return las

    def get_info(self):
        d = self.endpoint.get("/ept.json")
        info = Info(d)
        return info

    def count(self):
        loop = asyncio.get_event_loop()
        o = loop.run_until_complete(self.overlaps())
        return sum(k.count for k in self.overlaps_dict)

    def data(self):
        loop = asyncio.get_event_loop()
        if not self.overlaps_dict:
            o = loop.run_until_complete(self.overlaps())
        o = loop.run_until_complete(self.adata())
        return o

    async def adata(self):
        limit = 10
        connector = aiohttp.TCPConnector(limit=None)
        async with aiohttp.ClientSession(connector=connector) as session, TaskPool(
            limit
        ) as tasks:

            for key in self.overlaps_dict:
                url = "/ept-data/" + key.id() + ".laz"
                await tasks.put(self.endpoint.aget(url, session))

        laz = [LAZ(tasks.data[i]["result"]) for i in tasks.data]
        return laz

    async def overlaps(self):
        k = Key()
        k.coords = self.info.bounds

        f = "/ept-hierarchy/" + k.id() + ".json"

        async with aiohttp.ClientSession() as session:
            d = await self.endpoint.aget(f, session)
            hier = json.loads(d)
            await self._overlaps(self.endpoint, self.overlaps_dict, hier, k, session)

    async def _overlaps(self, endpoint, overlaps_dict, hier, key, session):

        if self.queryBounds:
            if not key.overlaps(self.queryBounds):
                return

        # if we have already set self.depthEnd
        # dont set it again
        if self.queryResolution and not self.computedDepth and not self.depthEnd:
            currentResolution = (
                self.info.bounds[3] - self.info.bounds[0]
            ) / self.info.span

            self.depthEnd = 1
            while currentResolution > self.queryResolution:
                currentResolution = currentResolution / 2.0
                self.depthEnd = self.depthEnd + 1
            self.computedDepth = True

        if self.depthEnd:
            if key.d >= self.depthEnd:
                return

        try:
            numPoints = hier[key.id()]
        except KeyError:
            hier[key.id()] = 0
            numPoints = 0
            return

        if numPoints == -1:
            # fetch more hierarchy

            f = "/ept-hierarchy/" + key.id() + ".json"
            data = await self.endpoint.aget(f, session)
            hier = json.loads(data)
            await self._overlaps(self.endpoint, self.overlaps_dict, hier, key, session)

        else:
            # check overlaps in each direction
            key.count = numPoints
            self.overlaps_dict[key] = key
            for direction in range(8):
                await self._overlaps(
                    self.endpoint,
                    self.overlaps_dict,
                    hier,
                    key.bisect(direction),
                    session,
                )



import numpy

class Schema(object):
    def __init__(self, data):
        self.dimensions = self._get_dimensions(data)

    def __len__(self):
        return len(self.dimensions)

    def _get_dimensions(self, data):
        dimensions = {}
        for d in data:
            name = d['name']

            kind = 'i'
            if d['type'] == 'unsigned':
                kind = 'u'
            elif d['type'] == 'signed':
                kind = 'i'
            elif d['type'] == 'float':
                kind = 'f'
            else:
                raise TypeError("Unrecognized type '%s'. Unable to convert to numpy dtype" % d['type'])

            d['dtype'] = kind+str(d['size'])

            dimensions[name] = d

        return dimensions


    def get_dtype(self):
        dt = []
        for d in self.dimensions:
            dim = self.dimensions[d]
            dt.append((dim['name'], dim['dtype']))

#        import pdb;pdb.set_trace()
        return numpy.dtype(dt)
    dtype = property(get_dtype)


